﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Configuration;
using System.Net;
using Mint.MintChip.Contract;
using Mint.MintChip.Api;

namespace MintXCore
{
    public class MintChip
    {
        private static ChipHandler chipHandler;
        private static string chipType;
        private static ChipHandler candidateLocalChipHandler, candidateRemoteChipHandler;

        public static ChipHandler GetChipHandlerInstance(Configuration conf)
        {
            if (chipHandler == null)
            {
                if (!SetConfigValues(conf))
                    return null;
                if (chipType == "local")
                    chipHandler = new LocalChipHandler();
                else if (chipType == "remote")
                    chipHandler = new RemoteChipHandler();
                if (!chipHandler.Initialize(conf))
                {
                    chipHandler = null;
                }
            }
            return chipHandler;
        }

        public static ChipHandler PreemptivelyGetChipHandlerInstance(Configuration conf, string chipTypePossibility)
        {
            if (!SetConfigValues(conf, true))
                return null;
            if (chipTypePossibility.ToLowerInvariant() == "local")
            {
                candidateLocalChipHandler = new LocalChipHandler();
                if (!candidateLocalChipHandler.Initialize(conf))
                    return null;
                return candidateLocalChipHandler;
            }

            if (chipTypePossibility.ToLowerInvariant() == "remote")
            {
                candidateRemoteChipHandler = new RemoteChipHandler();
                if (!candidateRemoteChipHandler.Initialize(conf))
                    return null;
                return candidateRemoteChipHandler;
            }

            return null;
        }

        private static bool SetConfigValues(Configuration conf, bool ignoreMissingChipType = false)
        {
            try
            {
                chipType = conf.AppSettings.Settings["chipType"].Value.ToLowerInvariant();
                switch (chipType)
                {
                    case "local": break;
                    case "remote": string remote = conf.AppSettings.Settings["RemoteRCMServer"].Value;
                        new Uri(remote);
                        break;
                    default: throw new ArgumentOutOfRangeException();
                }
            }
            catch (NullReferenceException)
            {
                if (ignoreMissingChipType) return true;
                return false;
            }
            catch (ArgumentOutOfRangeException)
            {
                return false;
            }

            return true;
        }
    }

    public abstract class ChipHandler
    {
        
        internal IMintChip mintChip;
        internal IMintChipStatus status;
        internal static Configuration config;

        public abstract bool Initialize(Configuration configuration);

        public string RequestPayment(int amountInCents, string annotationMsg = "", string responseAddress = "")
        {
            string file = Path.Combine(Path.GetTempPath(), Path.GetRandomFileName());
            try
            {
                IValueRequestMessage request = MessageFactory.Instance.CreateValueRequestMessage(mintChip.Id, amountInCents, mintChip.CurrencyCode);
                request.Annotation = annotationMsg;
                request.Challenge = Convert.ToInt32(annotationMsg.GetHashCode() + DateTime.UtcNow.GetHashCode());
                request.ResponseAddress = responseAddress;
                using (StreamWriter outfile = new StreamWriter(file))
                {
                    outfile.Write(request.ToBase64String());
                }
            }
            catch (MintChipException m)
            {
                //Console.WriteLine(m.Message);
                //return string.Empty;
                return m.Message;
            }
            return file;
        }

        public string GetMintChipID()
        {
            return mintChip.Id;
        }

        public string GetChipBalance()
        {
            return mintChip.GetStatus().Balance.ToMoney();
        }

        public string GetOfferredAmount(string eFile, bool lookForCredits)
        {
            string amount = string.Empty;            
            try
            {
                string valueStr = string.Empty;
                using (StreamReader infile = new StreamReader(eFile))
                {
                    valueStr = infile.ReadToEnd();
                }
                if (lookForCredits)
                {
                    IValueMessage valueMessage = (IValueMessage)MessageFactory.Instance.ToMessage(valueStr);
                    amount = valueMessage.Amount.ToMoney(); // no other checks necessary, since they have already passed
                    if (valueMessage.Amount == 1)
                    {
                        if (valueMessage.Annotation.StartsWith("COUNTER_OFFER_CENTS:"))
                        {
                            int cents = -1;
                            string money = valueMessage.Annotation.Substring("COUNTER_OFFER_CENTS:".Length);
                            int.TryParse(money, out cents);
                            return "Offer of " + cents.ToMoney(); // check YayAcceptAndThanksButton_Click
                        }
                    }
                }
                else
                {
                    IValueRequestMessage request = (IValueRequestMessage)MessageFactory.Instance.ToMessage(valueStr);
                    amount = request.Amount.ToMoney();
                }
            }
            catch (MintChipException m)
            {
                if (!string.Equals(m.Message, config.AppSettings.Settings["DuplicatePaymentError"].Value, StringComparison.InvariantCultureIgnoreCase))
                {
                    amount = "Unable to evaluate payment: " + m.Message;
                }
            }
            catch (IOException e)
            {
                amount = "Unable to accept evaluate: " + e.Message;
            }
            return amount;
        }


        public void PopulatePaymentDataFromRequest(MintChipMailData data)
        {
            string eFile = data.attachmentFilePath;
            string payeeId = string.Empty;
            int challenge = 0;
            try
            {
                string valueStr = string.Empty;
                using (StreamReader infile = new StreamReader(eFile))
                {
                    valueStr = infile.ReadToEnd();
                }
                IValueRequestMessage request = (IValueRequestMessage)MessageFactory.Instance.ToMessage(valueStr);
                payeeId = request.PayeeId;
                challenge = request.Challenge;
            }
            catch (MintChipException)
            {
            }
            catch (IOException)
            {
            }
            data.payeeId = payeeId;
            data.challenge = challenge;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="ecnFile"></param>
        /// <returns>0 if already paid, -1 in case this is not meant for this chip or the actual amount in cents</returns>
        public int GetAmountInCentsFromValueMessage(string eFile, bool lookForCredits)
        {
            int amount = 0;
            if (!File.Exists(eFile))
                return -1;
            try
            {
                string valueStr = string.Empty;
                using (StreamReader infile = new StreamReader(eFile))
                {
                    valueStr = infile.ReadToEnd();
                }
                
                if (lookForCredits)
                {
                    IValueMessage valueMessage = (IValueMessage)MessageFactory.Instance.ToMessage(valueStr);
                    if (string.Equals(valueMessage.PayeeId, this.mintChip.Id))
                    {
                        amount = valueMessage.Amount;
                        List<int> challengesInCreditedEntries = mintChip.ReadTransactionLog(LogType.CREDIT, 0, ushort.MaxValue).Select(x => x.Challenge).ToList();
                        if (challengesInCreditedEntries.Contains(valueMessage.Challenge))
                        {
                            amount = 0; // marker for already accepted payment
                            if (valueMessage.Annotation.StartsWith("COUNTER_OFFER_CENTS:"))
                            {
                                amount = -3; // special marker for already read mail with haggling offer.
                            }
                        }
                        else if ((amount == 1) && (valueMessage.Annotation.StartsWith("COUNTER_OFFER_CENTS:")))
                        {
                            amount = -2; // special marker for haggling
                        }
                    }
                    else
                    {
                        amount = -1; // marker for a mail not meant for the selected chip
                    }
                }
                else
                {
                    IValueRequestMessage request = (IValueRequestMessage)MessageFactory.Instance.ToMessage(valueStr);
                    if (!string.Equals(request.PayeeId, this.mintChip.Id))
                    {
                        amount = request.Amount;
                        List<int> challengesInDebitedEntries = (mintChip.ReadTransactionLog(LogType.DEBIT, 0, status.DebitLogCount)).Select(x => x.Challenge).ToList();
                        if (challengesInDebitedEntries.Contains(request.Challenge))
                            amount = 0;
                    }
                    else
                    {
                        amount = -1;
                    }
                }
            }
            catch (MintChipException m)
            {
                if (!string.Equals(m.Message, config.AppSettings.Settings["DuplicatePaymentError"].Value, StringComparison.InvariantCultureIgnoreCase))
                {
                    amount = -1;
                }
                else
                {
                    amount = 0;
                }
            }
            catch (IOException)
            {
                amount = -1;
            }
            return amount;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="erqFile"></param>
        /// <returns>0 if the amount has been paid before, -1 if this file is not meant for this chip and the actual amount in cents otherwise</returns>
        //public int GetRequestedAmountInCents(string erqFile)
        //{
        //    int amount = 0;
        //    if (!File.Exists(erqFile))
        //        return -1;
        //    try
        //    {
        //        string valueStr = string.Empty;
        //        using (StreamReader infile = new StreamReader(erqFile))
        //        {
        //            valueStr = infile.ReadToEnd();
        //        }
        //        IValueMessage valueMessage = null;
        //        valueMessage = (IValueMessage)MessageFactory.Instance.ToMessage(valueStr);
        //        if (!string.Equals(valueMessage.PayeeId, this.mintChip.Id))
        //        {
        //            amount = valueMessage.Amount;
        //            List<int> challengesInDebitedEntries = (mintChip.ReadTransactionLog(LogType.DEBIT, 0, status.DebitLogCount)).Select(x => x.Challenge).ToList();
        //            if (challengesInDebitedEntries.Contains(valueMessage.Challenge))
        //                amount = 0;
        //        }
        //        else
        //        {
        //            amount = -1;
        //        }
        //    }
        //    catch (MintChipException m)
        //    {
        //        if (!string.Equals(m.Message, config.AppSettings.Settings["DuplicatePaymentError"].Value, StringComparison.InvariantCultureIgnoreCase))
        //        {
        //            amount = -1;
        //        }
        //        else
        //        {
        //            amount = 0;
        //        }
        //    }
        //    catch (IOException e)
        //    {
        //        amount = -1;
        //    }
        //    //IValueRequestMessage valueRequestMessage = ShowRequestedPayment(erqFile);
        //    //if (valueRequestMessage == null) 
        //    //    amount = -1;
        //    //else
        //    //{
        //    //    List<int> challengesInDebitedEntries = (mintChip.ReadTransactionLog(LogType.DEBIT, 0, status.DebitLogCount)).Select(x => x.Challenge).ToList();
        //    //    if (!challengesInDebitedEntries.Contains(valueRequestMessage.Challenge))
        //    //    {
        //    //        amount = valueRequestMessage.Amount;
        //    //    }
        //    //}
        //    return amount;
        //}

        public List<Transactions> GetTransactions(int number)
        {
            List<ILogEntry> credits = new List<ILogEntry>(mintChip.ReadTransactionLog(LogType.CREDIT, 0, number));
            List<ILogEntry> debits = new List<ILogEntry>(mintChip.ReadTransactionLog(LogType.DEBIT, 0, number));
            credits.Reverse();
            debits.Reverse();
            List<Transactions> transactions = new List<Transactions>();
            int balance = mintChip.GetStatus().Balance;
            for (int i = 0, ci = 0, di = 0; i < number; i++)
            {
                if ((ci == credits.Count) && (di == debits.Count))
                    continue;
                Transactions t = null;
                if (ci == credits.Count)
                {
                    t = new Transactions(debits[di].TransactionTime, debits[di].Amount, balance, debits[di].PayeeId, Transactions.Type.Debit);
                    balance += debits[di].Amount;
                    di++;
                }
                else if (di == debits.Count)
                {
                    t = new Transactions(credits[ci].TransactionTime, credits[ci].Amount, balance, credits[ci].PayerId, Transactions.Type.Credit);
                    balance -= credits[ci].Amount;
                    ci++;
                }
                else
                {
                    if (credits[ci].TransactionTime > debits[di].TransactionTime)
                    {
                        t = new Transactions(credits[ci].TransactionTime, credits[ci].Amount, balance, credits[ci].PayerId, Transactions.Type.Credit);
                        balance -= credits[ci].Amount;
                        ci++;
                    }
                    else
                    {
                        t = new Transactions(debits[di].TransactionTime, debits[di].Amount, balance, debits[di].PayeeId, Transactions.Type.Debit);
                        balance += debits[di].Amount;
                        di++;
                    }
                }
                transactions.Add(t);
            }
            return transactions;
        }

        public List<Transactions> GetTransactions(DateTime fromDate, DateTime toDate)
        {
            List<ILogEntry> credits = new List<ILogEntry>(mintChip.ReadTransactionLog(LogType.CREDIT, 0, ushort.MaxValue)).Where(x => (x.TransactionTime >= fromDate) && (x.TransactionTime <= toDate)).ToList();
            List<ILogEntry> debits = new List<ILogEntry>(mintChip.ReadTransactionLog(LogType.DEBIT, 0, ushort.MaxValue)).Where(x => (x.TransactionTime >= fromDate) && (x.TransactionTime <= toDate)).ToList();
            List<ILogEntry> all = new List<ILogEntry>();
            all.AddRange(credits);
            all.AddRange(debits);
            all.Sort(CompareByTransactionDate);
            int balance = mintChip.GetStatus().Balance;
            List<Transactions> transactions = new List<Transactions>();
            foreach (ILogEntry temp in all)
            {
                Transactions t = null;
                if (temp.LogType == LogType.CREDIT)
                {
                    t = new Transactions(temp.TransactionTime, temp.Amount, balance, temp.PayerId, Transactions.Type.Credit);
                    balance -= temp.Amount;
                }
                else
                {
                    t = new Transactions(temp.TransactionTime, temp.Amount, balance, temp.PayeeId, Transactions.Type.Debit);
                    balance += temp.Amount;
                }
                transactions.Add(t);
            }

            return transactions;
        }

        private static int CompareByTransactionDate(ILogEntry x, ILogEntry y)
        {
            if (x == null)
            {
                if (y == null)
                {
                    return 0;
                }
                else
                {
                    return 1;
                }
            }
            else
            {
                if (y == null)
                {
                    return -1;
                }
                else
                {
                    return y.TransactionTime.CompareTo(x.TransactionTime);
                }
            }
        }

        public bool isValidForThisChip(string file)
        {
            bool validForThisChip = false;
            if (!File.Exists(file))
                return validForThisChip;
            return validForThisChip;
        }

        public void PrintChipData()
        {
            if (mintChip == null)
                throw new ArgumentNullException("Mintchip");
            status = mintChip.GetStatus();
            Console.WriteLine(mintChip.Id);
            Console.WriteLine(status.Balance);
            Console.WriteLine(status.MaxDebitAllowed);
            Console.WriteLine(status.MaxCreditAllowed);
            Console.WriteLine(status.DebitLogCount.ToString());
            Console.WriteLine(status.CreditLogCount.ToString());
            Console.WriteLine(status.DebitLogCountRemaining.ToString());
            Console.WriteLine(status.CreditLogCountRemaining.ToString());
            Console.WriteLine(mintChip.Version);
        }

        private IValueRequestMessage ShowRequestedPayment(string file)
        {
            IValueRequestMessage valueRequestMessage = null;
            if (!File.Exists(file))
                return valueRequestMessage;
            string valueStr = string.Empty;
            using (StreamReader infile = new StreamReader(file))
            {
                valueStr = infile.ReadToEnd();
            }
            try
            {
                valueRequestMessage = (IValueRequestMessage)MessageFactory.Instance.ToMessage(valueStr);
                //List<int> challengesInDebitedEntries = (mintChip.ReadTransactionLog(LogType.DEBIT, 0, status.DebitLogCount)).Select(x => x.Challenge).ToList();
                //if (challengesInDebitedEntries.Contains(valueRequestMessage.Challenge))
                //    return null; // This request has been paid previously
                if (string.Equals(valueRequestMessage.PayeeId, this.mintChip.Id))
                {
                    // Console.WriteLine("Ignoring this request, since it came from this chip itself");
                    return null;
                }
            }
            catch (MintChipException)
            {
                //Console.WriteLine("Invalid message: " + valueStr);
                return null;
            }
            return valueRequestMessage;
        }

        //public string PayMoney(IValueRequestMessage valueRequestMessage)
        //{
        //    string file = Path.Combine(Path.GetTempPath(), Path.GetRandomFileName());
        //    try
        //    {
        //        IValueMessage valueMessage = mintChip.CreateValueMessage(valueRequestMessage);
        //        using (StreamWriter outfile = new StreamWriter(file))
        //        {
        //            outfile.Write(valueMessage.ToBase64String());
        //        }
        //    }
        //    catch (MintChipException) { file = string.Empty; }
        //    catch (System.Net.WebException) { file = string.Empty; }
        //    return file;
        //}

        public string PayMoney(int amountInCents, string payeeID, string annotationMsg, int challenge)
        {
            string file = Path.Combine(Path.GetTempPath(), Path.GetRandomFileName());
            try
            {
                IValueRequestMessage valueRequestMessage = MessageFactory.Instance.CreateValueRequestMessage(payeeID, amountInCents, mintChip.CurrencyCode);
                valueRequestMessage.Annotation = annotationMsg;
                valueRequestMessage.Challenge = challenge != 0 ? challenge : Convert.ToInt32(annotationMsg.GetHashCode() + DateTime.UtcNow.GetHashCode());
                IValueMessage valueMessage = mintChip.CreateValueMessage(valueRequestMessage);
                using (StreamWriter outfile = new StreamWriter(file))
                {
                    outfile.Write(valueMessage.ToBase64String());
                }
            }
            catch (MintChipException me) { file = me.Message; }
            catch (WebException we) { file = we.Message; }
            return file;
        }

        public string AcceptPayment(string file)
        {
            IValueMessage valueMessage = null;
            if (!File.Exists(file))
                return "The value message could not be located";
            string valueStr = string.Empty;
            using (StreamReader infile = new StreamReader(file))
            {
                valueStr = infile.ReadToEnd();
            }
            try
            {
                valueMessage = (IValueMessage)MessageFactory.Instance.ToMessage(valueStr);
                if (string.Equals(valueMessage.PayeeId, this.mintChip.Id))
                {
                    mintChip.LoadValueMessage(valueMessage);
                    return string.Empty;
                }
                else
                {
                    // Console.WriteLine("Ignoring this payment, since it is not meant for this chip");
                    return "Ignoring this payment, since it is not meant for this chip";
                }
            }
            catch (MintChipException m)
            {
                if (!string.Equals(m.Message, config.AppSettings.Settings["DuplicatePaymentError"].Value, StringComparison.InvariantCultureIgnoreCase))
                {
                    return "Unable to accept payment: " + m.Message;
                }
            }
            catch (IOException e)
            {
                return "Unable to accept payment: " + e.Message;
            }
            return string.Empty;
        }
    }
}
