﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using Mint.MintChip.Contract;
using System.Security.Cryptography.X509Certificates;
using System.Net;
using System.Net.Security;
using System.Web.Script.Serialization;
using System.IO;
using Mint.MintChip.Api;

namespace MintXCore
{
    public class RemoteMintChipStatus : IMintChipStatus
    {
        internal RemoteMintChipStatus() { }

        public int Balance { get; internal set; }
        public int CreditLogCountRemaining { get; internal set; }
        public int CreditLogCount { get; internal set; }
        public int DebitLogCountRemaining { get; internal set; }
        public int DebitLogCount { get; internal set; }
        public int MaxCreditAllowed { get; internal set; }
        public int MaxDebitAllowed { get; internal set; }
    }

    internal sealed class LogEntry : ILogEntry
    {
        internal LogEntry() { }

        public int Amount { get; internal set; }
        public int Challenge { get; internal set; }
        public int Index { get; internal set; }
        public LogType LogType { get; internal set; }
        public string PayeeId { get; internal set; }
        public string PayerId { get; internal set; }
        public DateTime TransactionTime { get; internal set; }
    }

    public class RemoteChip : IMintChip
    {
        private const string CertIssuer = "C=CA, O=Royal Canadian Mint, OU=Remote MintChip, CN=Remote MintChip Certificate Authority";
        private const string ValueMessageContentType = "application/vnd.scg.ecn-message";
        private const string RequestContentType = "application/vnd.scg.ecn-request";

        private X509Certificate2 _clientCertificate;
        private string _serverUrl;

        private string InfoUrl
        {
            get
            {
                return this._serverUrl + "mintchip/info/json";
            }
        }

        private string CreateValueUrl
        {
            get
            {
                return this._serverUrl + "mintchip/payments/request";
            }
        }

        private string LoadValueUrl
        {
            get
            {
                return this._serverUrl + "mintchip/receipts";
            }
        }

        private string LastDebitUrl
        {
            get
            {
                return this._serverUrl + "mintchip/payments/lastdebit";
            }
        }

        private string DebitLogUrl
        {
            get
            {
                return this._serverUrl + "mintchip/payments/{0}/{1}/json";
            }
        }

        private string CreditLogUrl
        {
            get
            {
                return this._serverUrl + "mintchip/receipts/{0}/{1}/json";
            }
        }

        public byte[] Certificate
        {
            get
            {
                return null;
            }
        }

        private string _id;
        public string Id
        {
            get
            {
                if (string.IsNullOrEmpty(_id))
                {
                    var info = this.GetInfo();
                    this.PopulateProperties(info);
                }
                return _id;
            }

            private set
            {
                _id = value;
            }
        }

        private string _version;
        public string Version
        {
            get
            {
                if (string.IsNullOrEmpty(_version))
                {
                    var info = this.GetInfo();
                    this.PopulateProperties(info);
                }
                return _version;
            }

            private set
            {
                _version = value;
            }
        }

        private CurrencyCode? _currencyCode;
        public CurrencyCode CurrencyCode
        {
            get
            {
                if (!_currencyCode.HasValue)
                {
                    var info = this.GetInfo();
                    this.PopulateProperties(info);
                }
                return _currencyCode.Value;
            }

            private set
            {
                _currencyCode = value;
            }
        }


        internal RemoteChip(X509Certificate2 clientCertificate, string serverUrl)
        {
            if (clientCertificate == null)
            {
                throw new ArgumentNullException("clientCertificate");
            }

            if (string.IsNullOrEmpty(serverUrl))
            {
                throw new ArgumentNullException("serverUrl");
            }

            ServicePointManager.ServerCertificateValidationCallback += new RemoteCertificateValidationCallback(delegate
                (object sender, X509Certificate cert, X509Chain chain, SslPolicyErrors error)
            {
                return (cert.Issuer == CertIssuer);
            });

            this._serverUrl = serverUrl;
            if (!this._serverUrl.EndsWith("/") && !this._serverUrl.EndsWith(@"\"))
            {
                this._serverUrl += "/";
            }

            this._clientCertificate = clientCertificate;
        }


        public IValueMessage CreateValueMessage(IValueRequestMessage request)
        {
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }

            string response = this.ExecutePost(this.CreateValueUrl, request.ToBase64String(), RequestContentType);

            IValueMessage valueMesage = (IValueMessage)MessageFactory.Instance.ToMessage(response);

            return valueMesage;
        }


        public IValueMessage GetLastCreatedValueMessage(string annotation)
        {
            string response = this.ExecuteGet(this.LastDebitUrl);

            IValueMessage valueMesage = (IValueMessage)MessageFactory.Instance.ToMessage(response);

            return valueMesage;
        }


        public IMintChipStatus GetStatus()
        {
            var info = this.GetInfo();

            var status = new RemoteMintChipStatus
            {
                Balance = (int)info["balance"],
                CreditLogCount = (int)info["creditLogCount"],
                DebitLogCount = (int)info["debitLogCount"],
                CreditLogCountRemaining = (int)info["creditLogCountRemaining"],
                DebitLogCountRemaining = (int)info["debitLogCountRemaining"],
                MaxCreditAllowed = (int)info["maxCreditAllowed"],
                MaxDebitAllowed = (int)info["maxDebitAllowed"]
            };

            if (string.IsNullOrEmpty(this.Id))
            {
                this.PopulateProperties(info);
            }

            return status;
        }


        public bool IsValidId(string mintChipId)
        {
            System.Text.RegularExpressions.Regex r = new System.Text.RegularExpressions.Regex(@"^\d+$");
            System.Text.RegularExpressions.Match m = r.Match(mintChipId);
            return m.Success;
        }


        public void LoadValueMessage(IValueMessage valueMessage)
        {
            if (valueMessage == null)
            {
                throw new ArgumentNullException("valueMessage");
            }

            this.ExecutePost(this.LoadValueUrl, valueMessage.ToBase64String(), ValueMessageContentType);
        }


        public ILogEntry[] ReadTransactionLog(LogType logType, int startIndex, int numOfEntries)
        {
            if (startIndex < 0 || startIndex > Mint.MintChip.Core.MintChipData.MAX_LOG_COUNT)
            {
                throw new ArgumentException("Parameter 'startIndex' must be an integer less than " + Mint.MintChip.Core.MintChipData.MAX_LOG_COUNT + " and greater than 0.");
            }

            if (numOfEntries < 0 || (startIndex + numOfEntries) > Mint.MintChip.Core.MintChipData.MAX_LOG_COUNT)
            {
                throw new ArgumentException("Parameter 'numOfEntries' must be a positive integer less than " + (Mint.MintChip.Core.MintChipData.MAX_LOG_COUNT - startIndex));
            }

            if (numOfEntries == 0)
            {
                return new ILogEntry[0]; // If there are no entries, return an empty array right away
            }

            int stopIndex = startIndex + numOfEntries;

            var response = string.Empty;
            try
            {
                response = this.ExecuteGet(
                     string.Format((logType == LogType.CREDIT) ? this.CreditLogUrl : this.DebitLogUrl,
                     startIndex, stopIndex));
            }
            catch (MintChipException)
            {
                return new ILogEntry[0];
            }

            JavaScriptSerializer serializer = new JavaScriptSerializer();

            var jsonEntries = (object[])serializer.DeserializeObject(response);

            var entries = from entry in jsonEntries.Cast<Dictionary<string, object>>()
                          select new LogEntry
                          {
                              Amount = (int)entry["amount"],
                              Challenge = Convert.ToInt32(entry["challenge"].ToString(), 16),
                              Index = (int)entry["index"],
                              LogType = (LogType)entry["logType"],
                              PayerId = entry["payerId"] as string,
                              PayeeId = entry["payeeId"] as string,
                              TransactionTime = DateTime.Parse(entry["transactionTime"].ToString())
                          };

            return entries.ToArray();
        }


        #region "private methods"

        private void PopulateProperties(Dictionary<string, object> info)
        {
            this.Id = info["id"] as string;
            this.Version = info["version"] as string;
            this.CurrencyCode = (CurrencyCode)info["currencyCode"];
        }


        private Dictionary<string, object> GetInfo()
        {
            string response = this.ExecuteGet(this.InfoUrl);

            JavaScriptSerializer serializer = new JavaScriptSerializer();

            return serializer.DeserializeObject(response) as Dictionary<string, object>;
        }


        private string ExecuteGet(string url)
        {
            var request = (HttpWebRequest)WebRequest.Create(url);
            return ExecuteHttpRequest(request);
        }


        private string ExecutePost(string url, string postData = "", string mimeType = "")
        {
            var request = (HttpWebRequest)WebRequest.Create(url);
            request.Method = "POST";
            request.ContentLength = postData.Length;

            if (!string.IsNullOrEmpty(mimeType))
            {
                request.ContentType = mimeType;
            }

            if (!string.IsNullOrEmpty(postData))
            {
                byte[] data = Encoding.UTF8.GetBytes(postData);
                using (Stream requestStream = request.GetRequestStream())
                {
                    requestStream.Write(data, 0, data.Length);
                }
            }

            return ExecuteHttpRequest(request);
        }


        private string ExecuteHttpRequest(HttpWebRequest request)
        {
            request.ClientCertificates.Add(this._clientCertificate);

            string responseText = string.Empty;
            try
            {
                // hack to get the connection work with untrusted ssl certs
                // see http://dobrzanski.net/2009/04/01/systemnetwebexception-the-underlying-connection-was-closed-could-not-establish-trust-relationship-for-the-ssltls-secure-channel/
                ServicePointManager.ServerCertificateValidationCallback += (sender, certificate, chain, sslPolicyErrors) => true;
                using (var response = (HttpWebResponse)request.GetResponse())
                {
                    using (StreamReader reader = new StreamReader(response.GetResponseStream()))
                    {
                        responseText = reader.ReadToEnd();
                    }
                }
            }
            catch (WebException ex)
            {
                var response = (HttpWebResponse)ex.Response;
                if (response != null)
                {
                    throw new MintChipException(
                            string.Format("Failed to execute HttpWebRequest for the URL '{0}'. Status code: {1}. Status description: {2}.",
                                (request != null) ? request.RequestUri.ToString() : "", response.StatusCode.ToString(), response.StatusDescription),
                            ex);
                }
                else
                {
                    throw;
                }
            }
            finally
            {
                ServicePointManager.ServerCertificateValidationCallback -= (sender, certificate, chain, sslPolicyErrors) => true;
            }

            return responseText;
        }

        #endregion "private methods"
    }
}
